class Task:
    def __init__(self, name):
        self.name = name

    def configure(self):
        """Check if the machine is properly configured to run this task."""
        pass

    def run(self):
        """Run the task."""
        pass

    def complete(self):
        """Check if the task has completed successfully."""
        pass

    def log_progress(self):
        """Log the progress of the task."""
        pass


# import pynvml as pynvml
# import psutil

#? pynvml is much faster than nvidia-smi, but an additional dependency
#? Maybe should process this via https://gist.github.com/afspies/7e211b83ca5a8902849b05ded9a10696
import os
import subprocess
import time

# process_exceptions=['Xorg'], user_exceptions=[''], min_memory=6, base_on_memory=True, base_on_process=True
def check_gpu_usage(threshold_vram_usage=4000, max_gpus=2, wait=False, sleep_time=10):
    """
    Assigns free gpus to the current process via the CUDA_AVAILABLE_DEVICES env variable
    This function should be called after all imports,
    in case you are setting CUDA_AVAILABLE_DEVICES elsewhere
    Borrowed and fixed from https://gist.github.com/afspies/7e211b83ca5a8902849b05ded9a10696
    Args:
        threshold_vram_usage (int, optional): A GPU is considered free if the vram usage is below the threshold
                                              Defaults to 4000 (MiB).
        max_gpus (int, optional): Max GPUs is the maximum number of gpus to assign.
                                  Defaults to 2.
        wait (bool, optional): Whether to wait until a GPU is free. Default False.
        sleep_time (int, optional): Sleep time (in seconds) to wait before checking GPUs, if wait=True. Default 10.
    """

    def _check():
        # Get the list of GPUs via nvidia-smi
        smi_query_result = subprocess.check_output(
            "nvidia-smi -q -d Memory | grep -A4 GPU", shell=True
        )
        # Extract the usage information
        gpu_info = smi_query_result.decode("utf-8").split("\n")
        gpu_info = list(filter(lambda info: "Used" in info, gpu_info))
        gpu_info = [
            int(x.split(":")[1].replace("MiB", "").strip()) for x in gpu_info
        ]  # Remove garbage
        # Keep gpus under threshold only
        free_gpus = [
            str(i) for i, mem in enumerate(gpu_info) if mem < threshold_vram_usage
        ]
        free_gpus = free_gpus[: min(max_gpus, len(free_gpus))]
        gpus_to_use = ",".join(free_gpus)
        return gpus_to_use

    # while True:
        # gpus_to_use = _check()
        # if gpus_to_use or not wait:
            # break
        # print(f"No free GPUs found, retrying in {sleep_time}s")
        # time.sleep(sleep_time)

    gpus_to_use = _check()
    print(f"[[GPU INFO]] [{gpus_to_use}] Free")

# def check_gpu_usage(process_exceptions=['Xorg'], user_exceptions=[''], min_memory=6, base_on_memory=True, base_on_process=True):
#     # Process exceptions -> ignore these procs
#     # User exceptions -> avoid GPUs if proc running from these users
#     pynvml.nvmlInit()
#     deviceCount = pynvml.nvmlDeviceGetCount()
#     free_gpus = []
#     for i in range(deviceCount):

#         handle = pynvml.nvmlDeviceGetHandleByIndex(i)
        
#         mem = pynvml.nvmlDeviceGetMemoryInfo(handle)
#         free_memory = mem.free/(1024**3)
#         if base_on_memory and free_memory < min_memory:
#             continue

#         free = True 
#         if base_on_process:
#             procs = [*pynvml.nvmlDeviceGetComputeRunningProcesses(handle), *pynvml.nvmlDeviceGetGraphicsRunningProcesses(handle)]
#             for p in procs:
#                 try:
#                     process = psutil.Process(p.pid)
#                 except psutil.NoSuchProcess:
#                     continue

#                 if process.name not in process_exceptions and process.username() in user_exceptions:
#                     free = False
#                     break
#         if free:
#             free_gpus.append(str(i))

#     print(f"[[GPU INFO]] [{','.join(free_gpus)}] Free")
#     pynvml.nvmlShutdown()


check_gpu_usage()


# # ----------------------------------------------------------------------------------------------------------------------
# import pynvml as pynvml
# import psutil
# def check_gpu_usage():
#     pynvml.nvmlInit()
#     print ("Driver Version:", pynvml.nvmlSystemGetDriverVersion())
#     deviceCount = pynvml.nvmlDeviceGetCount()
#     for i in range(deviceCount):
#         handle = pynvml.nvmlDeviceGetHandleByIndex(i)
#         mem = pynvml.nvmlDeviceGetMemoryInfo(handle)
#         print ("Device", i, ":", pynvml.nvmlDeviceGetName(handle), [bytes2human(x) for x in [mem.total, mem.free, mem.used]])

#         procs = [*pynvml.nvmlDeviceGetComputeRunningProcesses(handle), *pynvml.nvmlDeviceGetGraphicsRunningProcesses(handle)]
#         for p in procs:
#             # print(p.pid, p.usedGpuMemory if isinstance(p.usedGpuMemory, int) else -1)
#             process = psutil.Process(p.pid)
#             print(process.name(), process.username())
#     pynvml.nvmlShutdown()
# ----------------------------------------------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------------------------------------------
# Asychronous timeout iterator
# Taken from  https://stackoverflow.com/questions/50241696/how-to-iterate-over-an-asynchronous-iterator-with-a-timeout
from typing import *
import asyncio
T = TypeVar('T')
# async generator, needs python 3.6
async def timeout_iterator(it: AsyncIterator[T], timeo: float, sentinel: T) -> AsyncGenerator[T, None]:
    try:
        nxt = asyncio.ensure_future(it.__anext__())
        while True:
            try:
                yield await asyncio.wait_for(asyncio.shield(nxt), timeo)
                nxt = asyncio.ensure_future(it.__anext__())
            except asyncio.TimeoutError:
                yield sentinel
    except StopAsyncIteration:
        pass
    finally:
        nxt.cancel()  # in case we're getting cancelled our self
# ----------------------------------------------------------------------------------------------------------------------
import itertools
# Taken from https://stackoverflow.com/questions/5228158/cartesian-product-of-a-dictionary-of-lists
def dict_product(dicts):
    """
    >>> list(dict_product(dict(number=[1,2], character='ab')))
    [{'character': 'a', 'number': 1},
     {'character': 'a', 'number': 2},
     {'character': 'b', 'number': 1},
     {'character': 'b', 'number': 2}]
    """
    return (dict(zip(dicts, x)) for x in itertools.product(*dicts.values()))
# ----------------------------------------------------------------------------------------------------------------------
# Random Stuff
import os
from pathlib import Path
from contextlib import contextmanager
# Taken from https://stackoverflow.com/questions/41742317/how-can-i-change-directory-with-python-pathlib
@contextmanager
def set_directory(path: Path):
    # Sets the cwd within the contex
    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)

from datetime import datetime
from time import time
def get_time_diff(start_time):
    # It took me almost 30 minutes to get this working
    # I'm a bloody idiot 
    # Expects starttime to be a datetime time
    if isinstance(start_time, float):
        start_time = datetime.fromtimestamp(start_time)
    dt = datetime.fromtimestamp(time()) - start_time
    hours, remainder = divmod(dt.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    f = ''
    if hours > 0:
        f += f'{hours} Hours, '
    if minutes > 0:
        f += f'{minutes} Minutes '
    if  (hours + minutes) > 0:
        f += 'and '
    f += f'{seconds} Seconds'
    return f 

# Taken from https://github.com/XuehaiPan/nvitop/blob/main/nvitop/core/utils.py
KiB = 1 << 10
MiB = 1 << 20
GiB = 1 << 30
TiB = 1 << 40
PiB = 1 << 50
NA = -1
def bytes2human(x):  # pylint: disable=too-many-return-statements
    if x is None or x == NA:
        return NA

    if not isinstance(x, int):
        try:
            x = round(float(x))
        except ValueError:
            return NA

    if x < KiB:
        return f'{x}B'
    if x < MiB:
        return f'{round(x / KiB)}KiB'
    if x <= 20 * GiB:
        return f'{round(x / MiB)}MiB'
    if x < 100 * GiB:
        return '{:.2f}GiB'.format(round(x / GiB, 2))
    if x < 1000 * GiB:
        return '{:.1f}GiB'.format(round(x / GiB, 1))
    if x < 100 * TiB:
        return '{:.2f}TiB'.format(round(x / TiB, 2))
    if x < 1000 * TiB:
        return '{:.1f}TiB'.format(round(x / TiB, 1))
    if x < 100 * PiB:
        return '{:.2f}PiB'.format(round(x / PiB, 2))
    return '{:.1f}PiB'.format(round(x / PiB, 1))
# ----------------------------------------------------------------------------------------------------------------------


from meinsweeper import MSLogger
import contextlib
import io
import unittest
import logging


class TestLocalLogger(unittest.TestCase):
    logger = MSLogger()
    # Redirect logger to another handler so we can capture outputs
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.ERROR)
    logger.python_logger.addHandler(stream_handler)

    def test_train_loss(self):
        # Capture logging stream "INFO" in a string buffer
        with io.StringIO() as buf:
            # run the tests
            self.stream_handler.stream = buf
            self.logger.log_loss(0.0001, mode='train')
            self.assertEqual(buf.getvalue(), '[[LOG_ACCURACY TRAIN]] Step: 0; Losses: Train: 0.0001\n')
            self.stream_handler.flush()

    def test_validation_loss(self):
        # Capture logging stream "INFO" in a string buffer
        with io.StringIO() as buf:
            # run the tests
            self.stream_handler.stream = buf
            self.logger.log_loss(0.1, mode='val')
            self.assertEqual(buf.getvalue(), '[[LOG_ACCURACY VAL]] Step: 0; Losses: Val: 0.1\n')
            self.stream_handler.flush()

class TestLocalNode(unittest.TestCase):
    def test_node(self):
        pass


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
import asyncio

class ComputeNode(ABC):
    def __init__(self, log_q: asyncio.Queue) -> None:
        pass

    @abstractmethod
    def run(self):
        pass
    
    # @abstractmethod
    # def put(self, src, target, recursive=True):
    #     pass

    # @abstractmethod
    # def get(self, src, target, recursive=True):
    #     pass




from meinsweeper import run_sweep

targets = []  # Machines
USERNAME = 'afs219'
SSH_KEY_PATH = '/media/home/alex/.ssh/doc'

for i in range(1, 20):
    targets.append(
        {
            'type': 'ssh',
            'params': {
                'address': f'ray{i:02}.doc.ic.ac.uk',
                'username': USERNAME,
                'key_path': SSH_KEY_PATH
            }
        }
    )
for i in range(1, 27):
    # if i not in [18]: # Kitty GPUs
    targets.append(
        {
            'type': 'ssh',
            'params': {
                'address': f'gpu{i:02}.doc.ic.ac.uk',
                'username': USERNAME,
                'key_path': SSH_KEY_PATH
            }
        }
    )

run_configurations = []
cmd = 'python -c "import time;[print(f\'[[LOG_ACCURACY TRAIN]] Step: {i+1}; Losses: Train:0.1\') for i in range(10) if time.sleep(3) is None]"'
# print(cmd)
for i in range(200):
    run_configurations.append((cmd, str(i)))

steps = 10
run_sweep(run_configurations, targets, steps=steps)

#   hydra.run.dir=\\"{path}\\" \
#   +run_base_dir=\\"{path}\\" \
#   debug={DEBUG} \
#   dataset_path={dataset_path} \
#   {stringify_stages(stages)} \
#   {' '.join([f'{k}={parse_yaml_value(v)}' for k, v in sweep_params.items()])}\
#   {' '.join([f'{k}={parse_yaml_value(v)}' for k, v in cfg['hydra_overrides'].items()])}
#     """

# def parse_yaml_value(val):
#     if val is None:
#         return 'null'
#     elif isinstance(val, str):
#         if '~' in val: # Tilde in path etc., not used as removal
#             return f"'{val}'"  # Escape the tilde
#         return f"'{val}'"
#     elif isinstance(val, list):
#         return f'[{",".join(map(parse_yaml_value, val))}]'
#     else:
#         return str(val)

# def dict_product(dicts):
#     return (dict(zip(dicts, x)) for x in itertools.product(*dicts.values()))

# def stringify_stages(stages):
#     return " ".join(stringify_recurse(stages, 'stages', []))

# def stringify_recurse(stages, prefix, output):
#     for key, value in stages.items():
#         if not isinstance(value, dict):
#             output.append(f"+{prefix}.{key}={parse_yaml_value(value)}".replace(" ", ''))
#         else:
#             stringify_recurse(value, f"{prefix}.{key}", output)
#     return output



import logging
import os
from pathlib import Path

LOG_DIR = Path(os.getenv("MEINSWEEPER_LOG_DIR", "./logs"))
LOGGING_ENABLED = os.getenv("MEINSWEEPER_LOGGING_ENABLED", "True").lower() == "true"

if LOGGING_ENABLED:
    # make sure the log, and log/nodes directories exist
    LOG_DIR.mkdir(exist_ok=True)
    (LOG_DIR/'nodes').mkdir(exist_ok=True)
else:
    print("FYI Meinsweeper Logging is disabled")


# A lazy way to disable all loggers without needing to check each logging call
class SilentFilter(logging.Filter):
    def __init__(self, logging_enabled=True):
        self.logging_enabled = logging_enabled

    def filter(self, record):
        # If the global flag is set to False, suppress the log record
        return self.logging_enabled  # True => Will be logged


# Create a custom filter instance
log_filter = SilentFilter(logging_enabled=LOGGING_ENABLED)

# -- Configure General Logging --
# Configure the logger
global_logger = logging.getLogger("Run Manager")
global_logger.setLevel(logging.INFO)

if LOGGING_ENABLED:
    # Create a FileHandler to append logs to a shared log file
    file_handler = logging.handlers.RotatingFileHandler(
        LOG_DIR / "run.log",  # Path to the log file
        maxBytes=50 * 1024 * 1024,  # Maximum log file size (50MB)
        backupCount=2,  # Keep at most 2 log files
    )
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s"
        )
    )

    # Add the FileHandler to the central logger
    # global_logger = logging.getLogger('Run Manager')
    global_logger.addHandler(file_handler)
global_logger.addFilter(log_filter)


# -- Decorator to log function calls to general log --
# Use function closure to capture logger
def log_function_call(func):
    def wrapper(*args, **kwargs):
        global_logger.info(f"Calling {func.__name__} with args: {args}, kwargs: {kwargs}")
        return func(*args, **kwargs)
    return wrapper

# -- Functions to initialize node-specific logs -- 
def init_node_logger(node_name):
    # Configure the thread-specific logger
    node_logger = logging.getLogger(f'node_{node_name}')
    node_logger.setLevel(logging.INFO)

    if LOGGING_ENABLED:
        # Create a unique log file for each thread
        log_filename = f"node_{node_name}.log"

        # Create a FileHandler for the thread's log file
        #! Add logic to keep log files corresponding to crashed runs and move them into
        #! A special folder
        file_handler = logging.handlers.RotatingFileHandler(
            LOG_DIR / "nodes" / log_filename,  # Path to the log file
            maxBytes=25 * 1024 * 1024,  # Maximum log file size (25MB)
            backupCount=2,  # Keep at most 2 log files
        )
        file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s"
            )
        )

        # Add the FileHandler to the thread-specific logger
        node_logger.addHandler(file_handler)
    node_logger.addFilter(log_filter)
    return node_logger


from abc import ABC
class Configurer(ABC):
    def __init__(self):
    
    def node_availability_test(self, node: Node):
        # Can use run method.. etc. from node class

        pass


# class Configuration():
#     def __init__(self) -> None:
#         self.test = True
#         self.fish = 7
#         self.frog = "yes_please"

#     def __repr__(self):
#         cfg_attr = [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("__")]
#         return " ".join([f"{attr}={getattr(self,str(attr))}" for attr in cfg_attr])

# c = Configuration()
# print(c)


from abc import ABC, abstractmethod
import socket
import asyncio
import asyncssh
from pathlib import Path
import logging
import xml.etree.ElementTree as ET

from meinsweeper.modules.helpers.debug_logging import global_logger
from .nodes import *
import textwrap

MINIMUM_VRAM = 8  # in GigaBytes
USAGE_CRITERION = 0.8  # percentage (float) or -1 => no processes other than xorg
MAX_PROCESSES = -1  # -1 => no limit, Otherwise number of processes = min(#nodes, #tbc_runs)
RUN_TIMEOUT = 600 #240  # in seconds
MAX_RETRIES = 3  #! removed from PQ after this - should readd after some interval

class RunManager(object):
    """ The RunManager spawns processes on the target nodes from the pool of available ones
    """
    def __init__(self, targets: dict, task_q: asyncio.Queue, log_q: asyncio.Queue) -> None:
        self.targets = targets
        self.max_proc = len(targets) if MAX_PROCESSES == -1 else MAX_PROCESSES
        self.log_q = log_q
        self.task_q = task_q
        self.running_proc = 0
        self.tasks = []

    async def start_run(self):
        # Create queue of available target nodes
        self.target_q = asyncio.PriorityQueue(
        )  #maxsize=MAX_PROCESSES) #! This should be num tasks, also check behaviour
        [await self.target_q.put(Target(target, retries=MAX_RETRIES)) for target in self.targets]

        # Spawn workers
        # ? Should this be a while loop
        [self.tasks.append(asyncio.create_task(self.spawn_worker())) for _ in range(self.max_proc)]
        await asyncio.gather(*self.tasks)

    # Function which connects to node and runs task
    async def spawn_worker(self) -> None:
        while not self.task_q.empty():
            connected = False
            target = None
            retries = None
            while not connected:
                target = await self.target_q.get()

                # initialize logger
                # Create target
                global_logger.info(f"Connecting to {target}")
                if target.details['type'] == 'ssh':
                    node = SSHNode(log_q=self.log_q, timeout=RUN_TIMEOUT, **target.details['params'])
                    connected = await node.open_connection()

                if not connected:
                    global_logger.warning(f"Failed to connect to {target}")
                    if target.not_failed_us():
                        global_logger.info(f"Retrying {target}")
                        target.retry()
                        await self.target_q.put(target)
                    await asyncio.sleep(5)

            # Run task once a connection was opened
            if self.task_q.empty():  # In case someone else was faster NOTE should instead timeout here
                global_logger.info(f"Task queue empty, closing connection to {target}")
                self.target_q.task_done()
                return

            # Get a task and run it
            #! Replace with configurer objectj
            # NOTE should add timeout here in case of async funkiness
            cfg, label = await self.task_q.get()  # Get next task from queue

            # cfg will be of form {'param1': ..., 'param2': ..., 'param3': ...}
            cmd = 'source ~/.bashrc &&' + cfg
            global_logger.info(f"Running task {cmd}, {label} on {target}")
            
            #! Replace with runner class
            success = await node.run(cmd, label)
            if success:
                global_logger.info(f"Task {label} completed successfully on {target}")
                await self.target_q.put(target)
            else:
                if target.not_failed_us():
                    global_logger.info(f"Task {label} failed on {target}, retrying")
                    target.retry()
                    await self.target_q.put(target)
                await self.task_q.put((cfg, label)) # Need to do task again
            self.task_q.task_done()
            self.target_q.task_done()


class Target(object):
    """
    A wrapper object for a target node which allows ranking in the PQ
    based on retries 
    
    TODO Extend to include other attributes like vram, etc. for ranking 
    """
    def __init__(self, target, retries=3) -> None:
        self.details = target
        self.retries = retries

    def get_target(self):
        return self.target

    def retry(self):
        self.retries -= 1

    def not_failed_us(self):
        return self.retries > 0

    def __lt__(self, other):
        #To work with PQ => Return True if we have MORE retries left
        if self.retries > other.retries:
            return True



from time import time as time_now
from .helpers.utils import get_time_diff

import logging
from logging import Logger

from rich.text import Text
from rich.panel import Panel
from rich.progress import Progress, BarColumn, TextColumn, TimeRemainingColumn
from rich.table import Table
import asyncio

TABLE_REFRESH_RATE = 3  # Hz

#-------------------------------------------------------
# This class will allow people to log in a fashion compliant with MeinSweeper
# RUNS ON REMOTE
class MSLogger():
    """Wrapper around a python logger which will take train/test 
    losses and print them in a format compatible with MeinSweeper"""
    def __init__(self):
        self.python_logger = logging.getLogger('meinsweeper_logger')
        self.python_logger.setLevel(logging.INFO)
        self.step = {'train': 0, 'val': 0, 'test': 0}

    def log_loss(self, loss: float, mode: str = 'train', step: int = None):
        assert mode in ['train', 'val', 'test'], f"Only modes 'train', 'val' and 'test' are supported"
        if step is None:
            step = self.step[mode]
            self.step[mode] += 1

        log_str = f'[[LOG_ACCURACY {mode.upper()}]] Step: {step}; Losses: {mode.capitalize()}: {loss}'
        self.python_logger.info(log_str)

    # def __getstate__(self):
    #     state = self.__dict__.copy()
    #     state["python_logger"] = None
    #     return state

    # def __setstate__(self, d):
    #     self.__dict__.update(d)  # I *think* this is a safe way to do it
    #     self.python_logger = logging.getLogger(
    #         'meinsweeper_logger'
    #     )


# ------------------------------------------------------
# This class will display logging info in a rich table
# RUNS ON HOST
class LogParser():
    def __init__(self, log_q: asyncio.PriorityQueue, steps: int = None) -> None:
        self.log_q = log_q
        self.table = DisplayTable('Training Progress', 'Info', steps=steps)
        self.display = self.table.table
        # can't get procs from table to count, as this includes failed runs
        self.start_time = time_now()

    async def start_logger(self):
        while True:
            i, addr, label = await self.log_q.get()
            self.table.update(i, addr, label)
            self.log_q.task_done()

    def complete(self, live_session):
        time_elapsed = get_time_diff(self.start_time)
        msg = f"🚀 Done! Completed {self.table.num_runs_completed} runs in {time_elapsed} 🚀"
        live_session.update(Panel(Text(msg, style="bold magenta", justify="center")))


# ------------------------------------------------------
class DisplayTable():
    def __init__(self, name: str, table_type: str, steps=None):
        self.name = name
        self.steps = steps
        self.table = Table(title=name)
        self.table.add_column("Progress", style="cyan", justify="right")
        self.num_runs_completed = 0
        self.progress_bars = Progress(
            TextColumn("[bold blue]{task.fields[name]}: {task.percentage:.0f}%"),
            BarColumn(),
            TimeRemainingColumn(),
            TextColumn(
                "[bold orange] Loss: {task.fields[loss_total]:.2e} Test_Acc {task.fields[test_acc]:.2e}"
            ),
        )
        self.table.add_row(Panel.fit(self.progress_bars, title=name, border_style="green", padding=(1, 1)))
        self.host_map = {}

    def update(self, progress: dict, host: str, label: str):
        progress, line = (
            progress  # progress is parsed line (i.e. losses or status) and line is raw
        )

        if progress is None:
            return
        progress['raw'] = str(progress)

        if host not in self.host_map:
            host_pid = self.add(host, label)
            progress['completed'] = 0
            progress['loss_total'] = 0
            progress['test_acc'] = 0
            self.progress_bars.update(host_pid, **progress)
        else:
            host_pid = self.host_map[host]
            failed = progress.get('status', '') == 'failed'
            if not failed:
                self.progress_bars.update(host_pid, **progress)
            else:
                self.progress_bars.update(host_pid, total=1, completed=1)

            if self.progress_bars.tasks[host_pid].finished:
                if not failed:
                    self.num_runs_completed += 1
                self.progress_bars.update(host_pid, visible=False)
                # self.progress_bars.tasks[host_pid].visible = False # self.progress_bars.remove_task(host_pid) #! doesn't work properly for some reason... (even if cleaning up pids)
                del self.host_map[host]

    def add(self, host: str, label: str):
        #! Representation should be generated elsewhere
        #! Also, need to handle stages
        # truncate label to 30 characters
        if len(label) > 30:
            label = label[:30] + (label[30:] and "...")
        pid = self.progress_bars.add_task(
            "", name=host.split('.')[0] + f' {label}', total=self.steps if self.steps else 100000
        )
        self.host_map[host] = pid
        return pid

    def get_num_procs(self):
        return len(self.progress_bars.tasks)

    def __str__(self):
        return str(self.table)


import socket
from pathlib import Path

import asyncssh

from meinsweeper.modules.helpers.debug_logging import init_node_logger
from meinsweeper.modules.helpers.utils import timeout_iterator

from .abstract import ComputeNode


class SSHNode(ComputeNode):
    def __init__(self, address, username, log_q, password=None, key_path=None, timeout=5):
        assert password or key_path, "Either password or key_path must be provided."
        self.log_q = log_q
        self.name = f'ssh-{address}'
        self.connection_info = {
            'address': address,
            'username': username,
            'password': password,
            'key_path': key_path,
            'timeout': timeout
        }
        self.RUN_TIMEOUT = timeout
        self.node_logger = init_node_logger(self.name)

    async def open_connection(self):
        address, username, key_path, timeout = self.connection_info['address'], self.connection_info[
            'username'], self.connection_info['key_path'], self.connection_info['timeout']
        try:
            self.node_logger.info(f"Connecting SSH to {address}")
            self.conn = await asyncssh.connect(
                address,
                username=username,
                known_hosts=None,
                login_timeout=timeout,
                encoding='utf-8',
                client_keys=[key_path] if key_path else None,
                term_type='bash'
            )
            #! this makes me sad
            self.node_logger.info(f"Connecting SCP to {address}")
            self.scp_conn = await asyncssh.connect(
                address,
                username=username,
                known_hosts=None,
                login_timeout=timeout,
                encoding='utf-8',
                client_keys=[key_path] if key_path else None
            )

        except asyncssh.PermissionDenied as auth_err:
            self.node_logger.warning(f"didn't conenct to {address} with {auth_err}")
            # raise ValueError('Authentication failed. Check user name and SSH key configuration.') from auth_err
            return False
        except asyncssh.DisconnectError as ssh_err:
            self.node_logger.warning(f"disconnected from {address} with {ssh_err}")
            # raise ValueError("No Session") from ssh_err
            return False
        except TimeoutError:
            self.node_logger.warning(f"timeout from {address}")
            return False
        except socket.gaierror as target_err:
            self.node_logger.warning(
                f"Could not connect to {address} with {target_err}"
            )
            return False
        except OSError as err:
            self.node_logger.warning(f"Got {err} when connecting to {address}")
            return False

        # Check if GPU is free
        self.free_gpus = await self.check_gpu_free()
        if self.free_gpus is None:
            return False

        return True

    async def check_gpu_free(self):
        self.node_logger.info(f"Checking GPU on {self.connection_info['address']}") 
        try:
            await asyncssh.scp(Path(__file__).parent.parent / 'check_gpu.py', (self.scp_conn, '/tmp/check_gpu.py'))
        except asyncssh.sftp.SFTPFailure:
            self.node_logger.warning(f"Could not copy check_gpu.py to {self.connection_info['address']}")
            return None

        gpus = await self.conn.run('python /tmp/check_gpu.py')
        gpus = gpus.stdout
        if 'No Module named' in gpus:
            self.node_logger.warning(f"Missing pynvml module on {self.connection_info['address']} - cannot check GPU")
            raise ValueError('Missing pynvml module on remote machine - cannot check GPU')

        if gpus.startswith('[[GPU INFO]]'):
            gpu_info = gpus[14:].split(']')[0]
            if gpu_info:
                self.node_logger.info(f"Found free GPU {gpu_info} on {self.connection_info['address']}")
                return gpu_info
        return None

    async def run(self, command, label):
        """ 
        Should be non-blocking and return whever a signal is recieved
        """
        async with self.conn.create_process(command) as proc: 
            self.node_logger.info(f"Running {command} on {self.connection_info['address']}")
            await self.log_q.put((({'status': 'running'}, 'running'), self.connection_info['address'], label))

            try:
                async for line in timeout_iterator(
                    proc.stdout.__aiter__(), self.RUN_TIMEOUT, "TIMEOUT"
                ):
                    self.node_logger.info(
                        f"Got STDIN line {line} from {self.connection_info['address']}"
                    )
                    if line == "TIMEOUT":
                        self.node_logger.warning(
                            f"Timeout on {self.connection_info['address']}"
                        )
                        await self.log_q.put(
                            (
                                ({"status": "failed"}, "failed"),
                                self.connection_info["address"],
                                label,
                            )
                        )
                        return False

                    parsed_line = self.parse_log_line(line)
                    if parsed_line == "FAILED":  # TODO Make this not a hack
                        self.node_logger.warning(
                            f"Failed (caught via parsed line) on {self.connection_info['address']}"
                        )
                        await self.log_q.put(
                            (
                                ({"status": "failed"}, "failed"),
                                self.connection_info["address"],
                                label,
                            )
                        )
                        return False
                    await self.log_q.put(
                        ((parsed_line, line), self.connection_info["address"], label)
                    )

                async for err_line in proc.stderr:
                    self.node_logger.info(
                        f"Got STDERR line {err_line} from {self.connection_info['address']}"
                    )
                    if err_line != "":
                        self.node_logger.warning(f"got error {err_line} for label")
                        await self.log_q.put(
                            (
                                ({"status": "failed"}, "failed"),
                                self.connection_info["address"],
                                label,
                            )
                        )
                        return False
            except asyncssh.misc.ConnectionLost as err:
                self.node_logger.warning(
                    f"Connection lost {err} on {self.connection_info['address']}"
                )
                self.log_q.put(
                    (
                        ({"status": "failed"}, "failed"),
                        self.connection_info["address"],
                        label,
                    )
                )
                return False
            except Exception as err:
                self.node_logger.warning(
                    f'Got {err} on {self.connection_info["address"]}'
                )
                self.log_q.put(
                    (
                        ({"status": "failed"}, "failed"),
                        self.connection_info["address"],
                        label,
                    )
                )
                return False
        return True

    """
    LOG Format Description
    Log lines should contain [[LOG_ACCURACY TRAIN]] or [[LOG_ACCURACY TEST]]
    Followed by 'sections' divided with ';' and key value pairs divided with ':' (A list of K:V pairs for losses)
    For example:
        [[LOG_ACCURACY TRAIN]] Step: 10; Losses: Train: 0.25, Val: 0.65
    """
    
    @staticmethod
    def parse_log_line(line):
        # print line to stdout in spite of rich table
        # logging.debug(line)
        # logging.debug("PARSING LINE")

        out = None
        #! FIX this can fail if the log line is part of an error message - we will try to parse
        #! but the f-strings won't have been evaluated (meaning we can't convert {step} or {loss} to float)
        if '[[LOG_ACCURACY TRAIN]]' in line:
            out = {}
            line = line.split('[[LOG_ACCURACY TRAIN]]')[1]
            line = line.split(';')
            for section in line:
                if 'Elapsed' in section:
                    continue
                elif "Losses" in section:
                    loss_terms = section.split('Losses:')[1].split(',')
                    for loss_term in loss_terms:
                        loss_name, loss_value = map(lambda x: x.strip(), loss_term.split(':'))
                        out[f'loss_total'] = float(loss_value)
                elif 'Step' in section:
                    out['completed'] = int(section.split('Step:')[1])
        elif '[[LOG_ACCURACY TEST]]' in line:  #TODO something is broken here
            line = line.split(':')[1]
            out = {'test_acc': float(line.strip())}
        elif 'error' in line or 'RuntimeError' in line or 'failed' in line or 'Killed' in line:
            out = 'FAILED'
        return out

    def scp_put(self, src, dst, recursive=True):
        self.scp.put(src, dst, recursive=recursive)

    def scp_get(self, src, dst, recursive=True):
        self.scp.get(src, dst, recursive=recursive)

    def __str__(self) -> str:
        return f'SSH Node {self.connection_info["address"]} with user {self.connection_info["username"]}'


import asyncio

from rich.live import Live

from .due import *

# global console # will be set in main
from .modules import LogParser, RunManager

# Change these according to your own configuration.

SSH_TIMEOUT = 10  #seconds


def run_sweep(cmd_list, targets, steps=None):
    # cmd_list should be a list of strings or list of tuples (str, str)
    if isinstance(cmd_list[0], str): # no labels included
        cmd_list = [(cmd, f'job {i}') for i,cmd in enumerate(cmd_list)]
    asyncio.run(main(cmd_list, targets, steps=steps))

async def main(cmd_list, targets, steps=None):
    sweep_q = asyncio.Queue()  # Workers take from - All sweep cfgs are enqueued here
    log_q = asyncio.Queue()  # Workers put into - All log cfgs are enqueued here

    [await sweep_q.put(sweep_run) for sweep_run in cmd_list]

    # Start the run
    run_mgr = RunManager(targets, sweep_q, log_q)
    logger = LogParser(log_q, steps=steps)

    with Live(logger.display, refresh_per_second=3) as live:
        runs = asyncio.create_task(run_mgr.start_run())
        logs = asyncio.create_task(logger.start_logger())

        await asyncio.gather(runs)
        await log_q.join()  # Implicitly awaits logger too

        logger.complete(live)  #This could be moved into logger, but requires keeping track of pids
        logs.cancel()

# Use duecredit (duecredit.org) to provide a citation to relevant work to
# be cited. This does nothing, unless the user has duecredit installed,
# And calls this with duecredit (as in `python -m duecredit script.py`):
due.cite(
    Doi("10.1167/13.9.30"),
    description="light-weight framework for running experiments on arbitrary compute nodes",
    tags=["experiment-management"],
    path='meinsweeper'
)
